@@ -5,13 +5,14 @@ from fs.base import FS
 from fs.info import Info
 from fs.subfs import SubFS
 from fs.zipfs import ZipFS
-from fs import path as fspath, errors, copy, walk
+from fs import path as fspath, errors, copy, walk, move
 import json
 import mimetypes
 from shutil import copyfileobj
 from collections import OrderedDict
 from pathvalidate import is_valid_filename
 import io
+from typing import Callable
 
 
 def fill_fs(fs: FS, d: dict):
@@ -56,6 +57,13 @@ def to_vuefinder_resource(
     }
 
 
+def fs_type(fs: FS) -> str:
+    unwrapped = fs
+    while hasattr(unwrapped, "_wrap_fs"):
+        unwrapped = unwrapped._wrap_fs
+    return f"{unwrapped.__class__.__module__}.{unwrapped.__class__.__qualname__}"
+
+
 class Adapter(object):
     def __init__(self, key: str, fs: FS):
         self.key = key
@@ -63,7 +71,12 @@ class Adapter(object):
 
 
 class VuefinderApp(object):
-    def __init__(self, enable_cors: bool = False, include_raw=False):
+    def __init__(
+        self,
+        enable_cors: bool = False,
+        include_raw=False,
+        fs_type: Callable[[FS], str] | None = None,
+    ):
         self.endpoints = {
             "GET:index": self._index,
             "GET:preview": self._preview,
@@ -75,6 +88,7 @@ class VuefinderApp(object):
             "POST:newfile": self._newfile,
             "POST:rename": self._rename,
             "POST:move": self._move,
+            "POST:copy": self._copy,
             "POST:delete": self._delete,
             "POST:upload": self._upload,
             "POST:archive": self._archive,
@@ -85,6 +99,7 @@ class VuefinderApp(object):
         self._adapters: dict[str, FS] = OrderedDict()
         self.enable_cors = enable_cors
         self.include_raw = include_raw
+        self.fs_type = fs_type
 
     def add_fs(self, key: str, fs: FS):
         self._adapters[key] = fs
@@ -99,15 +114,16 @@ class VuefinderApp(object):
 
     def _get_adapter(self, request: Request) -> Adapter:
         key = request.args.get("adapter")
-        return Adapter(key, self._adapters.get(key, self._default.fs))
-
-    def _get_storages(self):
-        return list(self._adapters.keys())
+        return (
+            Adapter(key, self._adapters[key])
+            if key in self._adapters
+            else self._default
+        )
 
     def _get_full_path(self, request: Request) -> str:
         return request.args.get("path", self._get_adapter(request).key + "://")
 
-    def _fs_path(self, path: str) -> str:
+    def _abspath(self, path: str) -> str:
         if ":/" in path:
             return fspath.abspath(path.split(":/")[1])
         return fspath.abspath(path)
@@ -115,7 +131,15 @@ class VuefinderApp(object):
     def delegate(self, request: Request) -> tuple[FS, str]:
         adapter = self._get_adapter(request)
         path = self._get_full_path(request)
-        return adapter.fs, self._fs_path(path)
+        return adapter.fs, self._abspath(path)
+
+    def _split_path(self, path: str, fallback_fs: FS = None) -> tuple[FS, str]:
+        """Splits the full path into filesystem and absolute path"""
+        fallback_fs = fallback_fs or self._default.fs
+        if ":/" not in path:
+            return fallback_fs, self._abspath(path)
+        key = path.split(":/")[0]
+        return self._adapters.get(key, None) or fallback_fs, self._abspath(path)
 
     def _index(self, request: Request, filter: Union[str, None] = None) -> Response:
         adapter = self._get_adapter(request)
@@ -130,7 +154,13 @@ class VuefinderApp(object):
         return json_response(
             {
                 "adapter": adapter.key,
-                "storages": self._get_storages(),
+                "storages": list(self._adapters.keys()),
+                "storage_info": {
+                    name: {
+                        "filesystem": self.fs_type(fs) if self.fs_type else fs_type(fs)
+                    }
+                    for name, fs in self._adapters.items()
+                },
                 "dirname": self._get_full_path(request),
                 "files": [
                     to_vuefinder_resource(adapter.key, path, info, self.include_raw)
@@ -201,7 +231,8 @@ class VuefinderApp(object):
     def _newfile(self, request: Request) -> Response:
         fs, path = self.delegate(request)
         name = request.get_json().get("name", "")
-        fs.writetext(fspath.join(path, name), "")
+        with fs.openbin(fspath.join(path, name), "wb"):
+            pass
         return self._index(request)
 
     def _rename(self, request: Request) -> Response:
@@ -212,28 +243,34 @@ class VuefinderApp(object):
         )
         return self._index(request)
 
-    def __move(self, fs, src, dst):
-        src = self._fs_path(src)
-        dst = self._fs_path(dst)
-        if fs.isdir(src):
-            fs.movedir(src, dst, create=True)
-        else:
-            fs.move(src, dst)
-
-    def _move(self, request: Request) -> Response:
+    def _transfer(
+        self, request: Request, transfer_dir: Callable, transfer_file: Callable
+    ) -> Response:
         fs, _ = self.delegate(request)
         payload = request.get_json()
-        dst_dir = payload.get("item", "")
+        dst_fs, dst_dir = self._split_path(payload.get("item", ""), fs)
         for item in payload.get("items", []):
-            src = item["path"]
-            self.__move(fs, src, fspath.combine(dst_dir, fspath.basename(src)))
+            src_fs, src_path = self._split_path(item["path"], fs)
+            dst_path = fspath.combine(dst_dir, fspath.basename(src_path))
+
+            if src_fs.isdir(src_path):
+                transfer_dir(src_fs, src_path, dst_fs, dst_path)
+            else:
+                transfer_file(src_fs, src_path, dst_fs, dst_path)
+
         return self._index(request)
 
+    def _move(self, request: Request) -> Response:
+        return self._transfer(request, move.move_dir, move.move_file)
+
+    def _copy(self, request: Request) -> Response:
+        return self._transfer(request, copy.copy_dir, copy.copy_file)
+
     def _delete(self, request: Request) -> Response:
         fs, path = self.delegate(request)
         payload = request.get_json()
         for item in payload.get("items", []):
-            path = self._fs_path(item["path"])
+            path = self._abspath(item["path"])
             if fs.isdir(path):
                 fs.removetree(path)
             else:
@@ -280,7 +317,7 @@ class VuefinderApp(object):
 
         fs, path = self.delegate(request)
         items: list[dict] = payload.get("items", [])
-        paths = [self._fs_path(item["path"]) for item in items if "path" in item]
+        paths = [self._abspath(item["path"]) for item in items if "path" in item]
         archive_path = fspath.join(path, name)
 
         if fs.exists(archive_path):
@@ -297,7 +334,7 @@ class VuefinderApp(object):
 
         fs, path = self.delegate(request)
         paths: list[str] = json.loads(request.args.get("paths", "[]"))
-        paths = [self._fs_path(path) for path in paths]
+        paths = [self._abspath(path) for path in paths]
 
         stream = io.BytesIO()
 
@@ -316,7 +353,7 @@ class VuefinderApp(object):
 
     def _unarchive(self, request: Request) -> Response:
         fs, path = self.delegate(request)
-        archive_path = self._fs_path(request.get_json().get("item"))
+        archive_path = self._abspath(request.get_json().get("item"))
 
         with fs.openbin(archive_path) as zip_file:
             with ZipFS(zip_file) as zip:
@@ -360,7 +397,7 @@ class VuefinderApp(object):
         response = None
         try:
             response = self.endpoints[endpoint](request)
-        except errors.ResourceReadOnly as exc:
+        except errors.FSError as exc:
             response = json_response({"message": str(exc), "status": False}, 400)
         except BadRequest as exc:
             response = json_response({"message": exc.description, "status": False}, 400)
